%% -*- erlang-indent-level: 2 -*-
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Copyright (c) 2001 by Erik Johansson.  All Rights Reserved 
%% Time-stamp: <01/04/10 14:20:15 happi>
%% ====================================================================
%%  Filename : 	hipe_rtl_arch.erl
%%  Module   :	hipe_rtl_arch
%%  Purpose  :  
%%  Notes    : 
%%  History  :	* 2001-04-10 Erik Johansson (happi@csd.uu.se): 
%%               Created.
%%  CVS      :
%%              $Author: kostis $
%%              $Date: 2001/08/16 10:48:49 $
%%              $Revision: 1.3 $
%% ====================================================================
%%  Exports  :
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

-module(hipe_rtl_arch).
-export([first_virtual_reg/0, 
	 proc_pointer_reg/0,
	 heap_pointer_reg/0,
	 heap_limit_reg/0,
	 fcalls_reg/0,
	 add_ra_reg/1,
	 reg_name/1,
	 is_precoloured/1,
	 live_at_return/0
	]).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% ____________________________________________________________________
%%      
%% ARCH-specific stuff
%% ____________________________________________________________________
%% 
%%
%% XXX: x86 might not _have_ real registers for some of these things
%%

first_virtual_reg() ->
    case get(hipe_target_arch) of
	ultrasparc ->
	    hipe_sparc_registers:first_virtual();
	x86 ->
	    hipe_x86_registers:first_virtual()
    end.

proc_pointer_reg() ->
    case get(hipe_target_arch) of
	ultrasparc ->
	    hipe_sparc_registers:proc_pointer();
	x86 ->
	    hipe_x86_registers:proc_pointer()
    end.

heap_pointer_reg() ->
    case get(hipe_target_arch) of
	ultrasparc ->
	    hipe_sparc_registers:heap_pointer();
	x86 ->
	    hipe_x86_registers:heap_pointer()
    end.

heap_limit_reg() ->
    case get(hipe_target_arch) of
	ultrasparc ->
	    hipe_sparc_registers:heap_limit();
	x86 ->
	    hipe_x86_registers:heap_limit()
    end.

fcalls_reg() ->
    case get(hipe_target_arch) of
	ultrasparc ->
	    hipe_sparc_registers:fcalls();
	x86 ->
	    hipe_x86_registers:fcalls()
    end.

add_ra_reg(Rest) ->
    case get(hipe_target_arch) of
	ultrasparc ->
	    [hipe_rtl:mk_var(hipe_sparc_registers:return_address()) | Rest];
	x86 ->
	    Rest
    end.

reg_name(Reg) ->
    case get(hipe_target_arch) of
	ultrasparc ->
	    hipe_sparc_registers:reg_name(Reg);
	x86 ->
	    hipe_x86_registers:reg_name(Reg)
    end.

is_precoloured(Reg) ->
    case get(hipe_target_arch) of
	ultrasparc ->
	    hipe_sparc_registers:is_precolored(Reg);
	x86 ->
	    hipe_x86_registers:is_precoloured(Reg)
    end.

live_at_return() ->
    case get(hipe_target_arch) of
	ultrasparc ->
	    ordsets:from_list(lists:map(fun(R) -> hipe_rtl:mk_reg(R) end,
					hipe_sparc_registers:global()));
	x86 ->
	    ordsets:from_list(lists:map(fun({R,_}) -> hipe_rtl:mk_reg(R) end,
					hipe_x86_registers:live_at_return()))
    end.
